@@ -1,3 +1,5 @@
+//! The app entry point.
+
 use super::common;
 use super::gui;
 use super::manager;
@@ -19,10 +21,9 @@ impl App {
         let (to_gui_loop_sender, to_gui_loop_receiver) = std::sync::mpsc::channel();
         self.to_gui_loop_receiver = Some(to_gui_loop_receiver);
 
-        let (from_gui_loop_sender,from_gui_loop_receiver) = std::sync::mpsc::channel();
+        let (from_gui_loop_sender, from_gui_loop_receiver) = std::sync::mpsc::channel();
         self.from_gui_loop_sender = Some(from_gui_loop_sender);
 
-
         std::thread::spawn(move || {
             let manager = manager::Manager::new(to_gui_loop_sender, from_gui_loop_receiver);
             f(manager);
@@ -37,14 +38,38 @@ impl App {
         };
         miniquad::start(conf, |mut ctx| {
             miniquad::UserData::owning(
-                gui::GuiLoop::new(&mut ctx, self.to_gui_loop_receiver.unwrap(), self.from_gui_loop_sender.unwrap()),
+                gui::GuiLoop::new(
+                    &mut ctx,
+                    self.to_gui_loop_receiver.unwrap(),
+                    self.from_gui_loop_sender.unwrap(),
+                ),
                 ctx,
             )
         });
     }
 }
 
+/// This spawns the main application thread - which one whishes to visually/interactively debug.
+///
+/// Example
+/// ``` no_run
+/// vviz::app::spawn(|mut manager: vviz::manager::Manager| {
+///     let mut ui_a_button = manager.add_button("a button".to_string());
+///
+///     // Some initial application logic...
+///
+///     loop {
+///        if ui_a_button.was_pressed() {
+///           println!("a button pressed");
+///         }
+///
+///         /// repeated application logic...
+///
+///         manager.sync_with_gui();
+///     }
+/// });
+/// ```
 pub fn spawn(f: fn(manager::Manager) -> ()) {
-    let my_gui = App::new();
-    my_gui.spawn(f)
-}
\ No newline at end of file
+    let vviz = App::new();
+    vviz.spawn(f)
+}
