@@ -4,6 +4,7 @@ use super::entities;
 use super::gui;
 
 use ::slice_of_array::prelude::*;
+use image::GenericImageView;
 
 /// Component such as a button or a slider.
 pub trait Component: downcast_rs::DowncastSync {
@@ -129,10 +130,8 @@ impl<T: Number> Component for Var<T> {
 pub struct RangedVar<T> {
     /// Current value.
     pub value: T,
-    /// Min bound.
-    pub min: T,
-    /// Max bound.
-    pub max: T,
+    /// Min, max bounds.
+    pub min_max: (T, T),
 }
 
 impl<T: Number> Component for RangedVar<T> {
@@ -143,7 +142,7 @@ impl<T: Number> Component for RangedVar<T> {
         sender: &mut std::sync::mpsc::Sender<Box<dyn FromGuiLoopMessage>>,
     ) {
         if ui
-            .add(egui::Slider::new(&mut self.value, self.min..=self.max).text(label))
+            .add(egui::Slider::new(&mut self.value, self.min_max.0..=self.min_max.1).text(label))
             .changed()
         {
             sender
@@ -213,11 +212,69 @@ mod offscreen_shader {
     }
 }
 
+/// [Widget] for 2d content.
+pub struct Widget2 {
+    aspect_ratio: f32,
+    maybe_image: Option<miniquad::Texture>,
+}
+
+impl Widget2 {
+    fn from_image(ctx: &mut miniquad::Context, image: image::DynamicImage) -> Self {
+        let rgba8 = image.to_rgba8();
+        let tex = miniquad::Texture::from_rgba8(
+            ctx,
+            rgba8.width() as u16,
+            rgba8.height() as u16,
+            rgba8.as_raw(),
+        );
+        Self {
+            aspect_ratio: image.dimensions().0 as f32 / image.dimensions().1 as f32,
+            maybe_image: Some(tex),
+        }
+    }
+
+    // fn from_aspect_ratio(aspect_ratio: f32) -> Self {
+    //     Self {
+    //         aspect_ratio,
+    //         maybe_image: None,
+    //     }
+    // }
+
+    // fn from_image_size(width: f32, height: f32) -> Self {
+    //     Self::from_aspect_ratio(width / height)
+    // }
+}
+
+impl Widget for Widget2 {
+    fn render(&mut self, _ctx: &mut miniquad::Context) {}
+
+    fn show(
+        &mut self,
+        ui: &mut egui::Ui,
+        assigned_width: f32,
+        assigned_height: f32,
+    ) -> Option<egui::Response> {
+        let w = (self.aspect_ratio * assigned_height).min(assigned_width);
+        let h = w / self.aspect_ratio;
+
+        let tex = egui::TextureId::User(self.maybe_image.unwrap().gl_internal_id() as u64);
+
+        let r = ui
+            .add(egui::Image::new(tex, egui::Vec2::new(w, h)).sense(egui::Sense::click_and_drag()));
+        Some(r)
+    }
+
+    fn aspect_ratio(&self) -> f32 {
+        self.aspect_ratio
+    }
+}
+
 /// [Widget] for 3d content such as meshes, line segments and point clouds.
 pub struct Widget3 {
+    camera_pose_scene: nalgebra::Isometry3<f32>,
     entities: linked_hash_map::LinkedHashMap<String, entities::NamedEntity3>,
-    offscreen_pipeline: miniquad::Pipeline,
-    //offscreen_bind: miniquad::Bindings,
+    mesh_pipeline: miniquad::Pipeline,
+    segments_pipeline: miniquad::Pipeline,
     offscreen_pass: miniquad::RenderPass,
     aspect_ratio: f32,
     texture_id: Option<egui::TextureId>,
@@ -254,7 +311,7 @@ impl Widget3 {
         )
         .unwrap();
 
-        let offscreen_pipeline = miniquad::Pipeline::with_params(
+        let mesh_pipeline = miniquad::Pipeline::with_params(
             ctx,
             &[miniquad::BufferLayout {
                 stride: (3 + 4) * std::mem::size_of::<f32>() as i32,
@@ -272,10 +329,33 @@ impl Widget3 {
             },
         );
 
+        let segments_pipeline = miniquad::Pipeline::with_params(
+            ctx,
+            &[miniquad::BufferLayout {
+                stride: (3 + 4) * std::mem::size_of::<f32>() as i32,
+                ..Default::default()
+            }],
+            &[
+                miniquad::VertexAttribute::new("pos", miniquad::VertexFormat::Float3),
+                miniquad::VertexAttribute::new("color0", miniquad::VertexFormat::Float4),
+            ],
+            offscreen_shader,
+            miniquad::PipelineParams {
+                depth_test: miniquad::Comparison::LessOrEqual,
+                depth_write: true,
+                primitive_type: miniquad::PrimitiveType::Lines,
+                ..Default::default()
+            },
+        );
+
         Self {
+            camera_pose_scene: nalgebra::Isometry3::<f32>::from_parts(
+                nalgebra::Translation3::<f32>::new(0.0, 0.0, -4.0),
+                nalgebra::UnitQuaternion::<f32>::from_euler_angles(0.0, 0.0, 0.),
+            ),
             entities: linked_hash_map::LinkedHashMap::new(),
-            offscreen_pipeline,
-            //offscreen_bind,
+            mesh_pipeline,
+            segments_pipeline,
             offscreen_pass,
             aspect_ratio: 640.0 / 480.0,
             texture_id: None,
@@ -287,13 +367,6 @@ impl Widget for Widget3 {
     fn render(&mut self, ctx: &mut miniquad::Context) {
         let proj = nalgebra_glm::perspective_fov_rh(60.0f32.to_radians(), 640.0, 480.0, 0.01, 10.0);
 
-        // TODO: implement mouse interaction / orbital control. For now the camera is placed at a
-        //       constant position looking at the scene.
-        let camera_pose_scene = nalgebra::Isometry3::<f32>::from_parts(
-            nalgebra::Translation3::<f32>::new(0.0, 0.0, -4.0),
-            nalgebra::UnitQuaternion::<f32>::from_euler_angles(0.0, 0.0, 0.),
-        );
-
         // the offscreen render pipeline, following this example:
         // https://github.com/not-fl3/egui-miniquad/blob/master/examples/render_to_egui_image.rs
         ctx.begin_pass(
@@ -301,41 +374,70 @@ impl Widget for Widget3 {
             miniquad::PassAction::clear_color(1.0, 1.0, 1.0, 1.),
         );
         for (_, named_entity) in &self.entities {
-            let vertex_buffer = miniquad::Buffer::immutable(
-                ctx,
-                miniquad::BufferType::VertexBuffer,
-                named_entity
-                    .entity
-                    .vertices
-                    .as_position_color()
-                    .unwrap()
-                    .vertices
-                    .flat(),
-            );
+            match &named_entity.entity {
+                entities::Entity3::Mesh(mesh) => {
+                    let vertex_buffer = miniquad::Buffer::immutable(
+                        ctx,
+                        miniquad::BufferType::VertexBuffer,
+                        mesh.vertices.as_position_color().unwrap().vertices.flat(),
+                    );
 
-            let index_buffer = miniquad::Buffer::immutable(
-                ctx,
-                miniquad::BufferType::IndexBuffer,
-                named_entity.entity.faces.indices.flat(),
-            );
+                    let index_buffer = miniquad::Buffer::immutable(
+                        ctx,
+                        miniquad::BufferType::IndexBuffer,
+                        mesh.faces.indices.flat(),
+                    );
 
-            let offscreen_bind = miniquad::Bindings {
-                vertex_buffers: vec![vertex_buffer],
-                index_buffer,
-                images: vec![],
-            };
+                    let offscreen_bind = miniquad::Bindings {
+                        vertex_buffers: vec![vertex_buffer],
+                        index_buffer,
+                        images: vec![],
+                    };
 
-            ctx.apply_pipeline(&self.offscreen_pipeline);
-            ctx.apply_bindings(&offscreen_bind);
+                    ctx.apply_pipeline(&self.mesh_pipeline);
+                    ctx.apply_bindings(&offscreen_bind);
 
-            let vs_params = offscreen_shader::Uniforms {
-                mvp: proj
-                    * camera_pose_scene.to_matrix()
-                    * named_entity.scene_pose_entity.to_matrix(),
-            };
-            ctx.apply_uniforms(&vs_params);
+                    let vs_params = offscreen_shader::Uniforms {
+                        mvp: proj
+                            * self.camera_pose_scene.to_matrix()
+                            * named_entity.scene_pose_entity.to_matrix(),
+                    };
+                    ctx.apply_uniforms(&vs_params);
 
-            ctx.draw(0, named_entity.entity.faces.indices.flat().len() as i32, 1);
+                    ctx.draw(0, mesh.faces.indices.flat().len() as i32, 1);
+                }
+                entities::Entity3::LineSegments(segments) => {
+                    let vertex_buffer = miniquad::Buffer::immutable(
+                        ctx,
+                        miniquad::BufferType::VertexBuffer,
+                        segments.vertices.vertices.flat(),
+                    );
+
+                    let index_buffer = miniquad::Buffer::immutable(
+                        ctx,
+                        miniquad::BufferType::IndexBuffer,
+                        segments.indices.flat(),
+                    );
+
+                    let offscreen_bind = miniquad::Bindings {
+                        vertex_buffers: vec![vertex_buffer],
+                        index_buffer,
+                        images: vec![],
+                    };
+
+                    ctx.apply_pipeline(&self.segments_pipeline);
+                    ctx.apply_bindings(&offscreen_bind);
+
+                    let vs_params = offscreen_shader::Uniforms {
+                        mvp: proj
+                            * self.camera_pose_scene.to_matrix()
+                            * named_entity.scene_pose_entity.to_matrix(),
+                    };
+                    ctx.apply_uniforms(&vs_params);
+
+                    ctx.draw(0, segments.indices.flat().len() as i32, 1);
+                }
+            }
         }
         ctx.end_render_pass();
 
@@ -364,6 +466,32 @@ impl Widget for Widget3 {
             egui::Image::new(self.texture_id.unwrap(), egui::Vec2::new(w, h))
                 .sense(egui::Sense::click_and_drag()),
         );
+
+        if ui.ctx().input().pointer.secondary_down() {
+            // TODO: Calculate delta scale based on scene depth.
+            let delta = 0.01 * ui.ctx().input().pointer.delta();
+            let mut scene_pose_camera = self.camera_pose_scene.inverse();
+            let translation_update = scene_pose_camera
+                .rotation
+                .transform_vector(&nalgebra::Vector3::new(-delta.x, -delta.y, 0.0));
+            scene_pose_camera.append_translation_mut(&nalgebra::Translation3 {
+                vector: translation_update,
+            });
+            self.camera_pose_scene = scene_pose_camera.inverse();
+        } else if ui.ctx().input().pointer.primary_down() {
+            // TODO: Rotates about scene center. Make the center point of rotation configurable.
+            let delta = 0.01 * ui.ctx().input().pointer.delta();
+            let mut scaled_axis = nalgebra::Vector3::zeros();
+            scaled_axis.x = -delta.y;
+            scaled_axis.y = delta.x;
+
+            let scene_rot_camera = self.camera_pose_scene.rotation.inverse();
+
+            self.camera_pose_scene.rotation *= nalgebra::UnitQuaternion::from_scaled_axis(
+                scene_rot_camera.transform_vector(&scaled_axis),
+            );
+        }
+
         Some(r)
     }
 
@@ -456,10 +584,8 @@ pub struct AddRangedVar<T> {
     pub label: String,
     /// Initial value.
     pub value: T,
-    /// Min bounds
-    pub min: T,
-    /// Max bounds
-    pub max: T,
+    /// Min, max bounds
+    pub min_max: (T, T),
 }
 
 impl<T: Number> ToGuiLoopMessage for AddRangedVar<T> {
@@ -468,13 +594,27 @@ impl<T: Number> ToGuiLoopMessage for AddRangedVar<T> {
             self.label,
             Box::new(RangedVar::<T> {
                 value: self.value,
-                min: self.min,
-                max: self.max,
+                min_max: self.min_max,
             }),
         );
     }
 }
 
+/// Adds [Widget2] to main panel.
+pub struct AddWidget2 {
+    /// Name of widget
+    pub label: String,
+    /// Image to show in (background of) widget
+    pub image: image::DynamicImage,
+}
+
+impl ToGuiLoopMessage for AddWidget2 {
+    fn update_gui(self: Box<Self>, data: &mut gui::GuiData, ctx: &mut miniquad::Context) {
+        data.widgets
+            .insert(self.label, Box::new(Widget2::from_image(ctx, self.image)));
+    }
+}
+
 /// Adds [Widget3] to main panel.
 pub struct AddWidget3 {
     /// Name of widget
